package de.fitko.fitconnect.api.services;

import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import de.fitko.fitconnect.api.domain.model.metadata.data.Data;
import de.fitko.fitconnect.api.domain.model.submission.Submission;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;

import java.util.Set;
import java.util.UUID;

/**
 * A technical system that accepts submissions on the administration side.
 *
 * @see <a href="https://docs.fitko.de/fit-connect/docs/receiving/overview">Receiving Submissions</a>
 */
public interface Subscriber {

    /**
     * Decrypts JWE-encrypted string data.
     *
     * @param privateKey       the private key to decrypt the JWE-encrypted string
     * @param encryptedContent JWE encrypted content that should be decrypted
     * @return the decrypted content as byte[]
     */
    byte[] decryptStringContent(final RSAKey privateKey, final String encryptedContent);

    /**
     * Polls available {@link SubmissionForPickup}s for a given destinationId.
     *
     * @param destinationId restricts the query to a specific destination
     * @param limit         number of submissions in result (max. is 500)
     * @param offset        position in the dataset
     * @return list of found {@link SubmissionForPickup}s
     */
    Set<SubmissionForPickup> pollAvailableSubmissions(UUID destinationId, int offset, int limit);

    /**
     * Gets a {@link Submission}.
     *
     * @param submissionId the unique identifier of a {@link Submission}
     * @return the requested {@link Submission}
     */
    Submission getSubmission(UUID submissionId);

    /**
     * Loads encrypted {@link Attachment} for a {@link Submission}.
     *
     * @param submissionId the unique identifier of a {@link Submission}
     * @param attachmentId unique identifier of the attachments
     * @return encrypted JWE string of attachment
     */
    String fetchAttachment(UUID submissionId, UUID attachmentId);

    /**
     * Validates the structure and signature of a Security-Event-Token (SET).
     *
     * @param caseId unique identifier of a {@link Submission}s case
     * @returna {@link ValidationResult}, contains an error if the SET is invalid
     */
    ValidationResult validateEventLog(UUID caseId);

    /**
     * Validates the {@link Metadata} structure against a given JSON-schema to ensure its correctness.
     *
     * @param metadata the {@link Metadata} object that is validated
     * @return a {@link ValidationResult}, contains an error if the {@link Metadata} is invalid or doesn't match the schema
     */
    ValidationResult validateMetadata(Metadata metadata, String schema);

    /**
     * Validates data integrity of {@link Attachment} or {@link Data}.
     *
     * @param originalHash original hash value of raw data as hex string
     * @param data unencrypted data as byte[]
     * @return a {@link ValidationResult}, contains an error if the hash values are not equal
     */
    ValidationResult validateHashIntegrity(String originalHash, byte[] data);

    /**
     * Sends a confirmation event if the received submission matches all validations.
     * @param submissionId unique identifier of submission
     * @param destinationId unique identifier of destination
     * @param caseId unique identifier of case
     * @see <a href="https://docs.fitko.de/fit-connect/docs/receiving/process-and-acknowledge">Process And Acknowledge</a>
     */
    void confirmValidSubmission(UUID submissionId, UUID destinationId, UUID caseId);

    /**
     * Sends a rejection event if the received submission violates any validation rule.
     *
     * @param submissionId unique identifier of submission
     * @param destinationId unique identifier of destination
     * @param caseId unique identifier of case
     * @see <a href="https://docs.fitko.de/fit-connect/docs/receiving/process-and-acknowledge">Process And Acknowledge</a>
     */
    void rejectSubmission(UUID submissionId, UUID destinationId, UUID caseId);
}
