package de.fitko.fitconnect.api.services.validation;

import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;

/**
 * Validator for publicKeys and metadata.
 *
 * @see <a href="https://docs.fitko.de/fit-connect/docs/sending/encrypt/#certificateValidation">Certificate Validation</a>
 */
public interface ValidationService {

    /**
     * Validates the public key consisting of the following steps:
     * <p>
     * <ul>
     *     <li>checks if the JSON Web Key is suitable for the encryption</li>
     *     <li>checks if the public key is matching the certificate referenced in the JWK</li>
     *     <li>checks the certificate chain up to the root certificate</li>
     *     <li>checks against a certificate revocation list and/or an OSCP-endpoint with signed response</li>
     * </ul>
     * @param publicKey the public JWK
     *
     * @return {@link ValidationResult} that includes an error if the validation failed
     */
    ValidationResult validatePublicKey(final RSAKey publicKey);

    /**
     * Validates the metadata against a given schema.
     *
     * @param metadata the current metadata object that should be validated
     * @param jsonSchema the schema the {@link Metadata} is validated against
     *
     * @return a {@link ValidationResult} with an optional error
     */
    ValidationResult validateMetadataSchema(final String jsonSchema, final Metadata metadata);


    /**
     * Compares a given byte[] to its original hash value.
     *
     * @param originalHexHash hash of the data as hex string
     * @param data byte[] of data that will be compared to the original hash
     * @return a {@link ValidationResult} with an optional error
     */
    ValidationResult validateIntegrity(String originalHexHash, byte[] data);

}
