package de.fitko.fitconnect.client;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentWithData;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.signature.Hash__1;
import de.fitko.fitconnect.api.domain.model.metadata.data.Data;
import de.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import de.fitko.fitconnect.api.domain.model.submission.Submission;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.exceptions.*;
import de.fitko.fitconnect.api.services.Subscriber;
import lombok.Builder;
import lombok.Getter;
import lombok.ToString;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.text.ParseException;
import java.util.*;

/**
 * A fluent client for handing in a subscription for a {@link Subscriber}
 */
public class SubscriberClient {

    private static final Logger logger = LoggerFactory.getLogger(SubscriberClient.class);


    private SubscriberClient() {
    }

    public static RequestSubmission builder(final Subscriber subscriber, final String privateKey, final String metadataSchema) {
        if (privateKey == null) {
            logger.error("No decryption key set, please check environment (config.yml) and provide a reference to the private decryption key");
            return null;
        }
        if (metadataSchema == null) {
            logger.error("No metadata schema present, please check environment (config.yml) and provide a metadata json schema");
            return null;
        }
        return new ClientBuilder(subscriber, privateKey, metadataSchema);
    }

    public interface RequestSubmission {

        /**
         * Loads a list of available {@link SubmissionForPickup} that were submitted to the {@link Subscriber}.
         *
         * @param destinationId unique identifier for a destination
         * @param limit         number of submissions in result (max. is 500)
         * @param offset        position in the dataset
         * @return set of unique of available submissions for pickup
         */
        Set<SubmissionForPickup> getAvailableSubmissions(UUID destinationId, int offset, int limit);

        /**
         * Loads a list of available {@link SubmissionForPickup} that were submitted to the {@link Subscriber}.
         *
         * @param destinationId unique identifier for a destination
         * @return set of unique available submissions for pickup
         */
        Set<SubmissionForPickup> getAvailableSubmissions(UUID destinationId);

        /**
         * Loads a single {@link SubmissionForPickup} by id.
         *
         * @param submissionId unique identifier of a {@link SubmissionForPickup}
         * @return next step to load availbale {@link Metadata}, {@link Data} and {@link Attachment}s
         */
        Optional<ReceivedSubmission> requestSubmission(UUID submissionId);
    }

    public static class ClientBuilder implements RequestSubmission {

        private static final int DEFAULT_SUBMISSION_LIMIT = 500;

        private final Subscriber subscriber;

        private final RSAKey decryptionKey;
        private final String metadataSchema;

        public ClientBuilder(final Subscriber subscriber, final String privateKey, final String metadataSchema) {
            this.subscriber = subscriber;
            this.decryptionKey = parseDecryptionKey(privateKey);
            this.metadataSchema = metadataSchema;
        }

        @Override
        public Set<SubmissionForPickup> getAvailableSubmissions(final UUID destinationId) {
            return getAvailableSubmissions(destinationId, 0, DEFAULT_SUBMISSION_LIMIT);
        }

        @Override
        public Set<SubmissionForPickup> getAvailableSubmissions(final UUID destinationId, final int offset, final int limit) {
            final Set<SubmissionForPickup> submissions = subscriber.pollAvailableSubmissions(destinationId, offset, limit);
            logger.info("Received {} submission(s) for destination {}", submissions.size(), destinationId);
            return submissions;
        }

        @Override
        public Optional<ReceivedSubmission> requestSubmission(final UUID submissionId) {
            try {

                if(decryptionKey == null) {
                    logger.error("Private decryption key is not available");
                    return Optional.empty();
                }

                /** Get submission by id **/
                final Submission submission = subscriber.getSubmission(submissionId);
                if (submission == null) {
                    logger.error("Submission with id {} does not exist", submissionId);
                    return Optional.empty();
                }

                /** Decrypt metadata and validate structure against schema **/
                final Metadata metadata = getMetadata(submission);
                final ValidationResult metadataValidation = subscriber.validateMetadata(metadata, metadataSchema);
                if (metadataValidation.hasError()) {
                    logger.error("Metadata does not match schema", metadataValidation.getError());
                    subscriber.rejectSubmission(submissionId, submission.getDestinationId(), submission.getCaseId());
                    return Optional.empty();
                }

                /** Decrypt data and validate hashes to ensure it was transferred without corruption **/
                final byte[] decryptedData = decryptData(submission.getEncryptedData());
                final String hashFromSender = getHashFromSender(metadata);
                final ValidationResult dataValidation = subscriber.validateHashIntegrity(hashFromSender, decryptedData);
                if (dataValidation.hasError()) {
                    logger.error("Data might be corrupted, hash-sum doe not match", dataValidation.getError());
                    subscriber.rejectSubmission(submissionId, submission.getDestinationId(), submission.getCaseId());
                    return Optional.empty();
                }

                /** Decrypt attachments and validate hashes to ensure they were transferred without corruption **/
                final List<Attachment> attachmentMetadata = metadata.getContentStructure().getAttachments();
                final List<AttachmentWithData> decryptedAttachments = decryptAttachments(submissionId, attachmentMetadata);
                final ValidationResult attachmentValidation = validateAttachments(decryptedAttachments);
                if (attachmentValidation.hasError()) {
                    logger.error("One of the attachments is invalid", attachmentValidation.getError());
                    subscriber.rejectSubmission(submissionId, submission.getDestinationId(), submission.getCaseId());
                    return Optional.empty();
                }

                /** Send CONFIRMED SET-event **/
                subscriber.confirmValidSubmission(submissionId, submission.getDestinationId(), submission.getCaseId());
                return Optional.of(buildReceivedSubmission(metadata, decryptedAttachments, decryptedData));

            } catch (final DecryptionException e) {
                logger.error("Decrypting payload failed", e);
            } catch (final RestApiException e) {
                logger.error("API request failed", e);
            } catch (final EventCreationException e) {
                logger.error("Creating SET event failed", e);
            } catch (final InvalidSigningKeyException  e) {
                logger.error("Signing the SET event failed", e);
            }catch (final IOException  e) {
                logger.error("Reading metadata failed", e);
            }
            return Optional.empty();
        }

        private List<AttachmentWithData> decryptAttachments(final UUID submissionId, final List<Attachment> attachmentMetadata) {
            if (attachmentMetadata == null || attachmentMetadata.isEmpty()) {
                logger.info("Submission contains no attachments");
                return Collections.emptyList();
            }
            return decryptAttachments(getAttachments(submissionId, attachmentMetadata));
        }

        private ReceivedSubmission buildReceivedSubmission(final Metadata metadata, final List<AttachmentWithData> decryptedAttachments, final byte[] decryptedData) {
            return ReceivedSubmission.builder()
                    .attachments(decryptedAttachments)
                    .data(new String(decryptedData, StandardCharsets.UTF_8))
                    .mimeType(metadata.getContentStructure().getData().getSubmissionSchema().getMimeType())
                    .build();
        }

        private RSAKey parseDecryptionKey(final String privateKey) {
            try {
                return RSAKey.parse(privateKey);
            } catch (final ParseException e) {
                logger.error("Private key could not be parsed", e);
                return null;
            }
        }

        private byte[] decryptData(final String encryptedData) {
            return subscriber.decryptStringContent(decryptionKey, encryptedData);
        }

        private Metadata getMetadata(final Submission submission) throws IOException {
            final byte[] metadataBytes = subscriber.decryptStringContent(decryptionKey, submission.getEncryptedMetadata());
            return new ObjectMapper().readValue(metadataBytes, Metadata.class);
        }

        private List<AttachmentWithData> getAttachments(final UUID submissionId, final List<Attachment> attachments) {
            final List<AttachmentWithData> attachmentsWithData = new ArrayList<>();
            for (final Attachment attachment : attachments) {
                final String encryptedAttachment = subscriber.fetchAttachment(submissionId, attachment.getAttachmentId());
                final AttachmentWithData attachmentWithData = AttachmentWithData.builder()
                        .encryptedData(encryptedAttachment)
                        .attachmentMetadata(attachment)
                        .build();
                attachmentsWithData.add(attachmentWithData);
            }
            return attachmentsWithData;
        }

        private List<AttachmentWithData> decryptAttachments(final List<AttachmentWithData> encryptedAttachments) {
            final List<AttachmentWithData> decryptedAttachments = new ArrayList<>();
            for (final AttachmentWithData encryptedAttachment : encryptedAttachments) {
                final byte[] decryptedData = subscriber.decryptStringContent(decryptionKey, encryptedAttachment.getEncryptedData());
                decryptedAttachments.add(encryptedAttachment.withDecryptedData(decryptedData));
            }
            return decryptedAttachments;
        }

        private ValidationResult validateAttachments(final List<AttachmentWithData> unencryptedAttachments) {
            for (final AttachmentWithData attachmentData : unencryptedAttachments) {
                final Attachment attachmentMetadata = attachmentData.getAttachmentMetadata();
                final Hash__1 hash = attachmentMetadata.getHash();
                final ValidationResult result = subscriber.validateHashIntegrity(hash.getContent(), attachmentData.getDecryptedData());
                if (result.hasError()) {
                    logger.error("Attachment data for id {} is corrupted", attachmentMetadata.getAttachmentId(), result.getError());
                    return ValidationResult.error(new DataIntegrityException("Attachment " + attachmentMetadata.getAttachmentId() + " is corrupt"));
                } else {
                    logger.info("Attachment {} with id {} is valid", attachmentMetadata.getFilename(), attachmentMetadata.getAttachmentId());
                }
            }
            return ValidationResult.ok();
        }

        private String getHashFromSender(final Metadata metadata) {
            return metadata.getContentStructure()
                    .getData()
                    .getHash()
                    .getContent();
        }
    }

    @Getter
    @Builder
    @ToString
    public static class ReceivedSubmission {
        String data;
        MimeType mimeType;
        List<AttachmentWithData> attachments;
    }
}