package de.fitko.fitconnect.client.cmd;

import com.beust.jcommander.JCommander;
import com.beust.jcommander.ParameterException;
import de.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentWithData;
import de.fitko.fitconnect.client.SenderClient;
import de.fitko.fitconnect.client.SubscriberClient;
import de.fitko.fitconnect.client.cmd.commands.ListAllSubmissionsCommand;
import de.fitko.fitconnect.client.cmd.commands.ListOneSubmissionCommand;
import de.fitko.fitconnect.client.cmd.commands.SendSubmissionCommand;
import de.fitko.fitconnect.client.cmd.util.AttachmentDataType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.stream.Collectors;

import static de.fitko.fitconnect.client.cmd.util.AttachmentDataType.JSON;
import static de.fitko.fitconnect.client.cmd.util.AttachmentDataType.XML;

public class CommandLineClient {

    private static final Logger logger = LoggerFactory.getLogger(CommandLineClient.class);

    private final SendSubmissionCommand sendSubmissionCommand;
    private final ListAllSubmissionsCommand listAllSubmissionsCommand;
    private final ListOneSubmissionCommand listOneSubmissionCommand;

    private final JCommander jc;

    private final SenderClient.WithDestination senderClient;
    private final SubscriberClient.RequestSubmission subscriberClient;

    public CommandLineClient(final SenderClient.WithDestination senderClient,
                             final SubscriberClient.RequestSubmission subscriberClient) {

        this.sendSubmissionCommand = new SendSubmissionCommand();
        this.listAllSubmissionsCommand = new ListAllSubmissionsCommand();
        this.listOneSubmissionCommand = new ListOneSubmissionCommand();

        this.senderClient = senderClient;
        this.subscriberClient = subscriberClient;
        jc = getJCommander();
    }

    public void run(final String... args) throws IOException {
        final var parsedCommand = parseCommandFromArguments(args);
        if (parsedCommand == null) {
            jc.usage();
            return;
        }
        executeCommand(parsedCommand);
    }

    private JCommander getJCommander() {
        return JCommander.newBuilder()
                .addCommand(sendSubmissionCommand)
                .addCommand(listAllSubmissionsCommand)
                .addCommand(listOneSubmissionCommand)
                .build();
    }

    private void executeCommand(final String parsedCommand) throws IOException {
        switch (parsedCommand) {
            case SendSubmissionCommand.SEND_CMD_NAME:
                sendSubmission(sendSubmissionCommand);
                break;
            case ListAllSubmissionsCommand.LIST_CMD_NAME:
                listSubmissions(listAllSubmissionsCommand);
                break;
            case ListOneSubmissionCommand.GET_CMD_NAME:
                listOneSubmission(listOneSubmissionCommand);
                break;
            default:
                jc.usage();
        }
    }

    private String parseCommandFromArguments(final String[] args) {
        try {
            jc.parse(args);
            return jc.getParsedCommand();
        } catch (final ParameterException e) {
            logger.error(e.getMessage(), "\n");
            jc.usage();
            return null;
        }
    }

    private void listOneSubmission(final ListOneSubmissionCommand listOneSubmissionCommand) throws IOException {
        logger.info("Listing submission for id {}", listOneSubmissionCommand.submissionId);
        final var submission = subscriberClient.requestSubmission(listOneSubmissionCommand.submissionId);
        if (submission.isEmpty()) {
            logger.info("No submission found for submission id {}", listOneSubmissionCommand.submissionId);
        } else {
            writeData(submission.get(), getTargetFolderPath(listOneSubmissionCommand));
        }
    }

    private void listSubmissions(final ListAllSubmissionsCommand listAllSubmissionsCommand) {
        logger.info("Listing available submissions for destination {}", listAllSubmissionsCommand.destinationId);
        subscriberClient.getAvailableSubmissions(listAllSubmissionsCommand.destinationId)
                .forEach(submission -> System.out.println(submission.getSubmissionId()));
    }

    private void sendSubmission(final SendSubmissionCommand sendSubmissionCommand) {
        logger.info("Sending new submission to destination {}", sendSubmissionCommand.destinationId);
        final var client = senderClient
                .withDestination(sendSubmissionCommand.destinationId)
                .withServiceType(sendSubmissionCommand.serviceName, sendSubmissionCommand.leikaKey)
                .withAttachments(sendSubmissionCommand.attachments.stream().map(File::new).collect(Collectors.toList()));
        if (sendSubmissionCommand.dataType == JSON) {
            client.withJsonData(sendSubmissionCommand.data).submit();
        } else if (sendSubmissionCommand.dataType == XML) {
            client.withXmlData(sendSubmissionCommand.data).submit();
        }
    }

    private String getTargetFolderPath(final ListOneSubmissionCommand listOneSubmissionCommand) {
        if (listOneSubmissionCommand.targetFolder != null) {
            return listOneSubmissionCommand.targetFolder + "/" + listOneSubmissionCommand.submissionId;
        } else {
            return listOneSubmissionCommand.submissionId.toString();
        }
    }

    private void writeData(final SubscriberClient.ReceivedSubmission submissionData, final String dataDirPath) throws IOException {
        logger.info("Creating data directory for submission in {}", dataDirPath);
        Files.createDirectories(Path.of(dataDirPath));

        final var fileEnding = AttachmentDataType.getFileTypeFromMimeType(submissionData.getMimeType());
        final var filePath = Path.of(dataDirPath + "/data." + fileEnding);
        logger.info("Writing data.{}", fileEnding);
        Files.write(filePath, submissionData.getData().getBytes(StandardCharsets.UTF_8));

        for (final AttachmentWithData attachment : submissionData.getAttachments()) {
            final String filename = attachment.getAttachmentMetadata().getFilename();
            logger.info("Writing attachment {}", filename);
            Files.write(Path.of(dataDirPath, filename), attachment.getDecryptedData());
        }
    }

}
