package de.fitko.fitconnect.client;

import de.fitko.fitconnect.api.config.*;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.client.factory.ClientFactory;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.condition.EnabledIfEnvironmentVariable;

import java.io.File;
import java.util.*;
import java.util.stream.Collectors;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.hasItems;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * This test uses real credentials and endpoints stored in gitlab ci variables to test the sdk against a real system
 */
class ClientIntegrationTest {

    final String authBaseUrl = "https://auth-testing.fit-connect.fitko.dev";
    final String submissionBaseUrl = "https://submission-api-testing.fit-connect.fitko.dev";

    @Test
    @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
    void testSendAndConfirmCycle() {

        // Given
        final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
        final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

        final var sentSubmission = ClientFactory.senderClient(config)
                .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withJsonData("{ data: 'Beispiel Fachdaten' }")
                .submit();

        assertTrue(sentSubmission.isPresent());

        // When
        final Optional<SubscriberClient.ReceivedSubmission> receivedSubmission =
                ClientFactory.subscriberClient(config)
                        .requestSubmission(sentSubmission.get().getSubmissionId());

        // Then
        assertTrue(receivedSubmission.isPresent());
        assertThat(receivedSubmission.get().getData(), is("{ data: 'Beispiel Fachdaten' }"));
        assertThat(receivedSubmission.get().getMimeType().toString(), is("application/json"));
        assertThat(new String(receivedSubmission.get().getAttachments().get(0).getDecryptedData()), is("Test attachment"));
    }

    @Test
    @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
    void testListSubmissions() {

        // Given
        final var testEnv = new Environment(authBaseUrl, submissionBaseUrl, true);
        final ApplicationConfig config = getConfigWithCredentialsFromGitlab(testEnv);

        final var senderClient = ClientFactory.senderClient(config);
        final var subscriberClient = ClientFactory.subscriberClient(config);

        final UUID destinationId = UUID.fromString(System.getenv("TEST_DESTINATION_ID"));
        final String leikaKey = "urn:de:fim:leika:leistung:99400048079000";
        final String serviceName = "Test Service";

        final var submissionOne = senderClient
                .withDestination(destinationId)
                .withServiceType(serviceName, leikaKey)
                .withJsonData("{ data: 'Beispiel Fachdaten 1' }")
                .submit();

        final var submissionTwo = senderClient
                .withDestination(destinationId)
                .withServiceType(serviceName, leikaKey)
                .withJsonData("{ data: 'Beispiel Fachdaten 2' }")
                .submit();

        assertTrue(submissionOne.isPresent());
        assertTrue(submissionTwo.isPresent());

        // When
        final Set<SubmissionForPickup> submissions = subscriberClient.getAvailableSubmissions(destinationId);

        // Then
        assertFalse(submissions.isEmpty());

        final List<UUID> submissionIds = submissions.stream().map(SubmissionForPickup::getSubmissionId).collect(Collectors.toList());

        assertThat(submissionIds, hasItems(submissionOne.get().getSubmissionId(), submissionTwo.get().getSubmissionId()));

        // remove by confirming
        subscriberClient.requestSubmission(submissionOne.get().getSubmissionId());
        subscriberClient.requestSubmission(submissionTwo.get().getSubmissionId());
    }

    @Test
    @EnabledIfEnvironmentVariable(named = "TEST_DESTINATION_ID", matches = ".*")
    void testAbortedSendSubmissionWithKeyValidationNotSilent() {

        // Given
        final var prodEnv = new Environment(authBaseUrl, submissionBaseUrl, false);
        final ApplicationConfig config = getConfigWithCredentialsFromGitlab("PROD", prodEnv);

        // When
        final var sentSubmission = ClientFactory.senderClient(config)
                .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                .withAttachment(new File("src/test/resources/attachment.txt"))
                .withJsonData("{ data: 'Beispiel Fachdaten' }")
                .submit();

        // Then
        assertTrue(sentSubmission.isEmpty());
    }

    private ApplicationConfig getConfigWithCredentialsFromGitlab(final Environment env) {
        return getConfigWithCredentialsFromGitlab("TESTING", env);
    }

    private ApplicationConfig getConfigWithCredentialsFromGitlab(final String environmentName, final Environment env) {

        final var sender = new SenderConfig(System.getenv("SENDER_CLIENT_ID"), System.getenv("SENDER_CLIENT_SECRET"));

        final var subscriber = SubscriberConfig.builder()
                .clientId(System.getenv("SUBSCRIBER_CLIENT_ID"))
                .clientSecret(System.getenv("SUBSCRIBER_CLIENT_SECRET"))
                .privateDecryptionKeyPath("src/test/resources/private_decryption_test_key.json")
                .build();

        final var resourcePaths = ResourcePaths.builder()
                .authTokenPath("/token")
                .destinationPath("/v1/destinations/{destinationId}")
                .destinationKeyPath("/v1/destinations/{destinationId}/keys/{kid}")
                .eventLogPath("/v1/cases/{caseId}/events")
                .submissionsPath("/v1/submissions")
                .submissionPath("/v1/submissions/{submissionId}")
                .submissionAttachmentPath("/v1/submissions/{submissionId}/attachments/{attachmentId}")
                .build();

        final var envName = new EnvironmentName(environmentName);

        return ApplicationConfig.builder()
                .senderConfig(sender)
                .subscriberConfig(subscriber)
                .environments(Map.of(envName, env))
                .resourcePaths(resourcePaths)
                .activeEnvironment(envName)
                .privateSigningKeyPath("src/test/resources/private_test_signing_key.json")
                .metadataSchemaPath("src/test/resources/metadata_schema.json")
                .build();
    }
}
