package de.fitko.fitconnect.core;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.model.destination.Destination;
import de.fitko.fitconnect.api.domain.model.jwk.JWK;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.exceptions.EncryptionException;
import de.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import de.fitko.fitconnect.api.exceptions.MetadataNotCreatedException;
import de.fitko.fitconnect.api.exceptions.RestApiException;
import de.fitko.fitconnect.api.services.Sender;
import de.fitko.fitconnect.api.services.crypto.CryptoService;
import de.fitko.fitconnect.api.services.destination.DestinationService;
import de.fitko.fitconnect.api.services.events.EventLogService;
import de.fitko.fitconnect.api.services.events.SecurityEventService;
import de.fitko.fitconnect.api.services.submission.SubmissionService;
import de.fitko.fitconnect.api.services.validation.ValidationService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.text.ParseException;
import java.util.UUID;

public class SubmissionSender implements Sender {

    private static final Logger logger = LoggerFactory.getLogger(SubmissionSender.class);

    private final ValidationService validationService;
    private final CryptoService cryptoService;
    private final DestinationService destinationService;
    private final SubmissionService submissionService;
    private final EventLogService eventLogService;
    private final SecurityEventService securityEventService;

    public SubmissionSender(final DestinationService destinationService,
                            final SubmissionService submissionService,
                            final EventLogService eventLogService,
                            final CryptoService encryptionService,
                            final ValidationService validationService,
                            final SecurityEventService securityEventService) {

        this.destinationService = destinationService;
        this.submissionService = submissionService;
        this.eventLogService = eventLogService;
        this.cryptoService = encryptionService;
        this.validationService = validationService;
        this.securityEventService =  securityEventService;
    }

    @Override
    public ValidationResult validatePublicKey(final RSAKey publicKey) {
        logger.info("Validating public key integrity");
        return validationService.validatePublicKey(publicKey);
    }

    @Override
    public ValidationResult validateMetadata(final Metadata metadata, final String schema) {
        return validationService.validateMetadataSchema(schema, metadata);
    }

    @Override
    public String encryptBytes(final RSAKey publicKey, final byte[] data) {
        logger.info("Encrypting ...");
        return cryptoService.encryptBytes(publicKey, data);
    }

    @Override
    public String createHash(final byte[] data) {
        logger.info("Creating hash ...");
        return cryptoService.hashBytes(data);
    }

    @Override
    public SubmissionForPickup createSubmission(final CreateSubmission submission) {
        logger.info("Creating new submission for destination {}", submission.getDestinationId());
        return submissionService.announceSubmission(submission);
    }

    @Override
    public void uploadAttachment(final UUID submissionId, final UUID attachmentId, final String encryptedAttachment) {
        logger.info("Uploading attachment(s) {} for submission {}", attachmentId, submissionId);
        submissionService.uploadAttachment(submissionId, attachmentId, encryptedAttachment);
    }

    @Override
    public RSAKey getEncryptionKeyForDestination(final UUID destinationId, final String encryptionKid) {
        logger.info("Getting encryption key for destination id {}", destinationId);
        try {
            final JWK jwk = destinationService.getEncryptionKey(destinationId, encryptionKid);
            return RSAKey.parse(new ObjectMapper().writeValueAsString(jwk));
        } catch (final ParseException | RestApiException | JsonProcessingException e) {
            throw new KeyNotRetrievedException("Parsing the encryption JWK failed", e);
        }
    }

    @Override
    public String encryptObject(final RSAKey encryptionKey, final Object obj) {
        try {
            return cryptoService.encryptString(encryptionKey, new ObjectMapper().writeValueAsString(obj));
        } catch (final JsonProcessingException | EncryptionException e) {
            throw new MetadataNotCreatedException("Failure creating metadata", e);
        }
    }

    @Override
    public void sendSubmission(final SubmitSubmission submission) {
        logger.info("Sending submission {}", submission.getSubmissionId());
        submissionService.sendSubmission(submission);
    }

    @Override
    public Destination getDestination(final UUID destinationId) {
        logger.info("Getting destination for id {}", destinationId);
        return destinationService.getDestination(destinationId);
    }

    @Override
    public void rejectSubmission(final UUID submissionId, final UUID destinationId, final UUID caseId) {
        final String rejectSubmissionEvent = securityEventService.createRejectSubmissionEvent(submissionId, destinationId, caseId);
        eventLogService.sendEvent(caseId, rejectSubmissionEvent);
        logger.info("REJECTED submission {}", submissionId);
    }

}
