package de.fitko.fitconnect.core;

import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.domain.model.metadata.Metadata;
import de.fitko.fitconnect.api.domain.model.submission.Submission;
import de.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import de.fitko.fitconnect.api.domain.validation.ValidationResult;
import de.fitko.fitconnect.api.services.Subscriber;
import de.fitko.fitconnect.api.services.crypto.CryptoService;
import de.fitko.fitconnect.api.services.events.EventLogService;
import de.fitko.fitconnect.api.services.events.SecurityEventService;
import de.fitko.fitconnect.api.services.submission.SubmissionService;
import de.fitko.fitconnect.api.services.validation.ValidationService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Set;
import java.util.UUID;

public class SubmissionSubscriber implements Subscriber {

    private static final Logger logger = LoggerFactory.getLogger(SubmissionSubscriber.class);

    private final SubmissionService submissionService;
    private final CryptoService cryptoService;
    private final ValidationService validationService;
    private final EventLogService eventLogService;
    private final SecurityEventService securityEventService;

    public SubmissionSubscriber(final SubmissionService submissionService,
                                final EventLogService eventLogService,
                                final CryptoService cryptoService,
                                final ValidationService validationService,
                                final SecurityEventService securityEventService) {
        this.submissionService = submissionService;
        this.eventLogService = eventLogService;
        this.cryptoService = cryptoService;
        this.validationService = validationService;
        this.securityEventService = securityEventService;
    }

    @Override
    public byte[] decryptStringContent(final RSAKey privateKey, final String encryptedContent) {
        return cryptoService.decryptBytes(privateKey, encryptedContent);
    }

    @Override
    public Set<SubmissionForPickup> pollAvailableSubmissions(final UUID destinationId, final int offset, final int limit) {
        return submissionService.pollAvailableSubmissions(destinationId, offset, limit).getSubmissions();
    }

    @Override
    public Submission getSubmission(final UUID submissionId) {
        return submissionService.getSubmission(submissionId);
    }

    @Override
    public String fetchAttachment(final UUID submissionId, final UUID attachmentId) {
        return submissionService.getAttachment(submissionId, attachmentId);
    }

    @Override
    public ValidationResult validateEventLog(final UUID caseId) {

        return ValidationResult.ok();
    }

    @Override
    public ValidationResult validateMetadata(final Metadata metadata, final String schema) {
        return validationService.validateMetadataSchema(schema, metadata);
    }

    @Override
    public ValidationResult validateHashIntegrity(final String originalHash, final byte[] data) {
        return validationService.validateIntegrity(originalHash, data);
    }

    @Override
    public void confirmValidSubmission(final UUID submissionID, final UUID destinationId, final UUID caseId) {
        final String confirmedSubmissionEvent = securityEventService.createAcceptSubmissionEvent(submissionID, destinationId,caseId);
        eventLogService.sendEvent(caseId, confirmedSubmissionEvent);
        logger.info("CONFIRMED submission {} successfully", submissionID);
    }

    @Override
    public void rejectSubmission(final UUID submissionId, final UUID destinationId, final UUID caseId) {
        final String rejectSubmissionEvent = securityEventService.createRejectSubmissionEvent(submissionId, destinationId, caseId);
        eventLogService.sendEvent(caseId, rejectSubmissionEvent);
        logger.info("REJECTED submission {}", submissionId);

    }

}
