package de.fitko.fitconnect.core.auth;

import de.fitko.fitconnect.api.domain.auth.OAuthToken;
import de.fitko.fitconnect.api.exceptions.AuthenticationException;
import de.fitko.fitconnect.api.exceptions.RestApiException;
import de.fitko.fitconnect.api.services.auth.OAuthService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import java.nio.charset.StandardCharsets;
import java.time.LocalDateTime;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;

import static de.fitko.fitconnect.core.util.Preconditions.checkNotNull;
import static java.util.stream.Collectors.joining;

public class DefaultOAuthService implements OAuthService {

    private static final Logger logger = LoggerFactory.getLogger(DefaultOAuthService.class);

    private final RestTemplate restTemplate;

    private final String tokenUrl;
    private final String clientId;
    private final String clientSecret;

    private OAuthToken currentToken;
    private LocalDateTime tokenExpirationTime;

    public DefaultOAuthService(final RestTemplate restTemplate, final String clientId, final String clientSecret, final String authUrl) {
        this.restTemplate = restTemplate;
        this.clientId = clientId;
        this.clientSecret = clientSecret;
        this.tokenUrl = authUrl;
        resetExistingToken();
    }

    @Override
    public synchronized OAuthToken getCurrentToken() throws RestApiException {
        if(tokenExpired()){
            logger.info("Current token is expired, authenticating ...");
            authenticate();
        }
        return currentToken;
    }

    private boolean tokenExpired() {
        if(currentToken == null || tokenExpirationTime == null){
            return true;
        }
        final var now = LocalDateTime.now();
        return tokenExpirationTime.isBefore(now) || tokenExpirationTime.isEqual(now);
    }

    private void resetExistingToken() {
        this.currentToken = null;
        this.tokenExpirationTime = null;
    }

    private void authenticate() throws AuthenticationException {
        checkNotNull("restTemplate", restTemplate);
        checkNotNull("clientId", clientId);
        checkNotNull("clientSecret", clientSecret);
        checkNotNull("tokenUrl", tokenUrl);
        final String requestBody = buildRequestBody(clientId, clientSecret);
        this.currentToken = performTokenRequest(requestBody);
        this.tokenExpirationTime = LocalDateTime.now().plusSeconds(this.currentToken.getExpiresIn());
    }

    private String buildRequestBody(final String clientId, final String clientSecret, final String... scope) {
        final var data = new HashMap<String, String>();
        data.put("grant_type", "client_credentials");
        data.put("client_id", clientId);
        data.put("client_secret", clientSecret);

        if (scope.length > 0) {
            data.put("scope", String.join(",", scope));
        }

        return joinRequestParams(data);
    }

    private String joinRequestParams(final HashMap<String, String> data) {
        return data.entrySet()
                .stream()
                .map(e -> e.getKey() + "=" + e.getValue())
                .collect(joining("&"));
    }

    private OAuthToken performTokenRequest(final String requestBody) throws RestApiException {
        final HttpHeaders headers = getHeaders();
        final HttpEntity<String> entity = new HttpEntity<>(requestBody, headers);
        try {
            logger.info("Sending authentication request");
            return restTemplate.exchange(tokenUrl, HttpMethod.POST, entity, OAuthToken.class).getBody();
        } catch (final RestClientException e) {
            throw new RestApiException("Could not retrieve OAuth token", e);
        }
    }

    private HttpHeaders getHeaders() {
        final var headers = new HttpHeaders();
        headers.setAccept(Collections.singletonList(MediaType.APPLICATION_JSON));
        headers.setContentType(MediaType.APPLICATION_FORM_URLENCODED);
        headers.setAcceptCharset(List.of(StandardCharsets.UTF_8));
        return headers;
    }

}
