package de.fitko.fitconnect.core.crypto;

import com.nimbusds.jose.*;
import com.nimbusds.jose.crypto.RSADecrypter;
import com.nimbusds.jose.crypto.RSAEncrypter;
import com.nimbusds.jose.jwk.RSAKey;
import de.fitko.fitconnect.api.exceptions.DecryptionException;
import de.fitko.fitconnect.api.exceptions.EncryptionException;
import de.fitko.fitconnect.api.services.crypto.CryptoService;
import de.fitko.fitconnect.api.services.crypto.MessageDigestService;

import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;
import java.security.NoSuchAlgorithmException;
import java.text.ParseException;

public class JWECryptoService implements CryptoService {

    public static final JWEAlgorithm ALGORITHM = JWEAlgorithm.RSA_OAEP_256;
    public static final EncryptionMethod ENCRYPTION_METHOD = EncryptionMethod.A256GCM;
    private final MessageDigestService messageDigestService;

    public JWECryptoService(final MessageDigestService messageDigestService) {
        this.messageDigestService = messageDigestService;
    }

    @Override
    public String decryptString(final RSAKey privateKey, final String encryptedData) throws DecryptionException {
        return decrypt(privateKey, encryptedData).toString();
    }

    @Override
    public byte[] decryptBytes(final RSAKey privateKey, final String encryptedData) throws DecryptionException {
        return decrypt(privateKey, encryptedData).toBytes();
    }

    @Override
    public String encryptString(final RSAKey publicKey, final String data) throws EncryptionException {
        final Payload payload = new Payload(data);
        return encrypt(publicKey, payload);
    }

    @Override
    public String encryptBytes(final RSAKey publicKey, final byte[] bytes) throws EncryptionException {
        final Payload payload = new Payload(bytes);
        return encrypt(publicKey, payload);
    }

    @Override
    public String hashBytes(final byte[] data) {
        final byte[] hash = messageDigestService.createHash(data);
        return messageDigestService.toHexString(hash);
    }

    private String encrypt(final RSAKey publicKey, final Payload payload) throws EncryptionException {
        try {
            final KeyGenerator keyGenerator = KeyGenerator.getInstance("AES");
            keyGenerator.init(ENCRYPTION_METHOD.cekBitLength());
            final SecretKey cek = keyGenerator.generateKey();
            final String keyID = getIdFromPublicKey(publicKey);
            return encryptPayload(publicKey, payload, cek, keyID);
        } catch (final NoSuchAlgorithmException | JOSEException e) {
            throw new EncryptionException(e.getMessage(), e);
        }
    }

    private Payload decrypt(final RSAKey privateKey, final String encData) throws DecryptionException {
        try {
            final JWEObject jwe = JWEObject.parse(encData);
            jwe.decrypt(new RSADecrypter(privateKey));
            return jwe.getPayload();
        } catch (final ParseException | JOSEException e) {
            throw new DecryptionException(e.getMessage(), e);
        }
    }

    private JWEObject getJWEObject(final String keyID, final Payload payload) {
        return new JWEObject(getJWEHeader(keyID), payload);
    }

    private JWEHeader getJWEHeader(final String keyID) {
        return new JWEHeader.Builder(ALGORITHM, ENCRYPTION_METHOD)
                .compressionAlgorithm(CompressionAlgorithm.DEF)
                .contentType("application/json")
                .keyID(keyID)
                .build();
    }

    private RSAEncrypter getEncrypter(final RSAKey publicKey, final SecretKey cek) {
        try {
            return new RSAEncrypter(publicKey.toRSAPublicKey(), cek);
        } catch (final JOSEException e) {
            throw new EncryptionException("RSAEncrypter could not be initialized", e);
        }
    }

    private String encryptPayload(final RSAKey publicKey, final Payload payload, final SecretKey cek, final String keyID) throws JOSEException, EncryptionException {
        final JWEObject jwe = getJWEObject(keyID, payload);
        jwe.encrypt(getEncrypter(publicKey, cek));
        checkIfJWEObjectIsEncrypted(jwe);
        return jwe.serialize();
    }

    private void checkIfJWEObjectIsEncrypted(final JWEObject jwe) throws EncryptionException {
        if (!jwe.getState().equals(JWEObject.State.ENCRYPTED)) {
            throw new EncryptionException("JWE object is not encrypted");
        }
    }

    private String getIdFromPublicKey(final RSAKey publicKey) throws EncryptionException {
        final String keyID = publicKey.getKeyID();
        if (keyID == null || keyID.isEmpty()) {
            throw new EncryptionException("public key has no keyID");
        }
        return keyID;
    }

}
