package de.fitko.fitconnect.core.auth;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import de.fitko.fitconnect.api.domain.auth.OAuthToken;
import de.fitko.fitconnect.api.services.auth.OAuthService;
import de.fitko.fitconnect.core.RestEndpointBase;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestTemplate;

import static com.github.tomakehurst.wiremock.client.WireMock.*;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertThrows;

class DefaultOAuthServiceTest extends RestEndpointBase {

    private OAuthService underTest;

    @BeforeEach
    public void setUp() {
        wireMockServer.resetMappings();
        final var config = getTestConfig("http://localhost:" +  wireMockServer.port());
        underTest = new DefaultOAuthService(new RestTemplate(), "id", "secret", config.getOAuthTokenEndpoint());
    }

    @Test
    void serviceNotInitialized() {
        // Given
        final OAuthService instance =  new DefaultOAuthService(null, null, null, null);

        // Then
        assertThrows(
                IllegalStateException.class,
                () -> instance.getCurrentToken(),
                "Expected getInstance() to throw, but nothing was thrown"
        );
    }

    @Test
    void getCurrentToken() throws JsonProcessingException {
        // Given
        final OAuthToken token= new OAuthToken();
        token.setAccessToken("123");
        token.setExpiresIn(30);

        stubForTokenPostRequest(token);

        // When
        final OAuthToken currentToken = underTest.getCurrentToken();

        // Then
        assertThat(currentToken.getExpiresIn(), is(30));
        assertThat(currentToken.getAccessToken(), is("123"));
    }

    @Test
    void refreshTokenWhenExpired() throws JsonProcessingException {
        // Given
        final OAuthToken oldToken = new OAuthToken();
        oldToken.setAccessToken("123");
        oldToken.setExpiresIn(0);

        // Set expired token so the next call needs to refresh
        stubForTokenPostRequest(oldToken);

        // Refresh
        wireMockServer.resetMappings();
        final OAuthToken refreshed = new OAuthToken();
        refreshed.setAccessToken("456");
        refreshed.setExpiresIn(1800);

        // When
        stubForTokenPostRequest(refreshed);
        final OAuthToken refreshedToken = underTest.getCurrentToken();

        // Then
        assertThat(refreshedToken.getExpiresIn(), is(1800));
        assertThat(refreshedToken.getAccessToken(), is("456"));
    }

    private void stubForTokenPostRequest(final OAuthToken token) throws JsonProcessingException {
        wireMockServer.stubFor(
                post(urlEqualTo("/token"))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(token))
                                .withHeader("Content-Type", MediaType.APPLICATION_JSON_VALUE)
                                .withStatus(200)));
    }
}