package de.fitko.fitconnect.core.submission;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.tomakehurst.wiremock.http.ResponseDefinition;
import de.fitko.fitconnect.api.config.*;
import de.fitko.fitconnect.api.domain.auth.OAuthToken;
import de.fitko.fitconnect.api.domain.model.submission.*;
import de.fitko.fitconnect.api.services.auth.OAuthService;
import de.fitko.fitconnect.api.services.submission.SubmissionService;
import de.fitko.fitconnect.core.RestEndpointBase;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestTemplate;

import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;

import static com.github.tomakehurst.wiremock.client.WireMock.*;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.when;

class SubmissionApiServiceTest extends RestEndpointBase {

    private RestTemplate restTemplate;
    private OAuthService authServiceMock;
    private SubmissionService underTest;

    @BeforeEach
    public void setUp() {

        final var fakeBaseUrl = "http://localhost:" + wireMockServer.port();
        final ApplicationConfig config = getTestConfig(fakeBaseUrl);

        restTemplate = new RestTemplate();
        authServiceMock = Mockito.mock(OAuthService.class);
        underTest = new SubmissionApiService(authServiceMock, restTemplate, config);
    }

    @Test
    void announceSubmission() throws JsonProcessingException {
        // Given
        final var caseId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final List<UUID> attachmentIdsToAnnounce = List.of(UUID.randomUUID());

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        final ServiceType serviceType = new ServiceType();
        serviceType.setName("FakeService");
        serviceType.setDescription("A faked service");
        serviceType.setIdentifier("de:fitco:fake:service");

        final var expectedSubmission = new SubmissionForPickup();
        expectedSubmission.setSubmissionId(UUID.randomUUID());
        expectedSubmission.setCaseId(UUID.randomUUID());
        expectedSubmission.setDestinationId(UUID.randomUUID());

        final var createSubmission = CreateSubmission.builder()
                .announcedAttachments(attachmentIdsToAnnounce)
                .destinationId(destinationId)
                .serviceType(serviceType)
                .caseId(caseId).build();

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        wireMockServer.stubFor(
                post(urlEqualTo("/v1/submissions"))
                        .withHeader("Authorization", containing("Bearer " + authToken.getAccessToken()))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedSubmission))
                                .withHeader("Content-Type", MediaType.APPLICATION_JSON_VALUE)
                                .withStatus(200)));

        // When
        final SubmissionForPickup announcedSubmission = underTest.announceSubmission(createSubmission);

        // Then
        assertNotNull(announcedSubmission);
        assertThat(announcedSubmission, is(expectedSubmission));
    }

    @Test
    void uploadAttachment() {

        // Given
        final var submissionId = UUID.randomUUID();
        final var attachmentId = UUID.randomUUID();
        final var encryptedAttachment = "eyJlbmMiOiJBMjU2R0NNIiwiYWxnIjoiUlNBLU9BRVAtMjU2In0.nlXGAufYH36IABDy0En0";


        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/submissions/" + submissionId + "/attachments/" + attachmentId;
        final ResponseDefinition response = wireMockServer.stubFor(
                put(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(encryptedAttachment)
                                .withHeader("Content-Type", "application/jose")
                                .withStatus(204))).getResponse();

        // When
        underTest.uploadAttachment(submissionId, attachmentId, encryptedAttachment);

        // Then
        verify(putRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/jose"));
        assertEquals(response.getStatus(), HttpStatus.NO_CONTENT.value());
    }

    @Test
    void getAttachment() {

        // Given
        final var submissionId = UUID.randomUUID();
        final var attachmentId = UUID.randomUUID();
        final var encryptedAttachment = "eyJlbmMiOiJBMjU2R0NNIiwiYWxnIjoiUlNBLU9BRVAtMjU2In0.nlXGAufYH36IABDy0En0";

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/submissions/" + submissionId + "/attachments/" + attachmentId;
        final ResponseDefinition response = wireMockServer.stubFor(
                get(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(encryptedAttachment)
                                .withHeader("Content-Type", "application/jose")
                                .withStatus(200))).getResponse();

        // When
        final String attachment = underTest.getAttachment(submissionId, attachmentId);

        // Then
        assertNotNull(attachment);
        assertThat(attachment, is(encryptedAttachment));
        verify(getRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/jose"));
        assertEquals(response.getStatus(), HttpStatus.OK.value());
    }

    @Test
    void sendSubmission() throws JsonProcessingException {

        // Given
        final var submissionId = UUID.randomUUID();
        final var encryptedMetadata = "GfC20IZSSchs27ADalHpRoTZKfXhc7hcMk8Y9V8yTP0jYbmrq6NtEg-QS2O5TQFD9";
        final var encryptedData = "31PBgKjPXHYX1Y6iUcR1sXxSUPjePi8F8P31PBgKjPXHYX1Y6iUcR1sXxSUPjePi8F8P";

        final SubmitSubmission submitSubmission = new SubmitSubmission();
        submitSubmission.setSubmissionId(submissionId);
        submitSubmission.setEncryptedMetadata(encryptedMetadata);
        submitSubmission.setEncryptedData(encryptedData);

        final Submission expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(submissionId);
        expectedSubmission.setEncryptedData(encryptedData);
        expectedSubmission.setEncryptedMetadata(encryptedMetadata);

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/submissions/" + submissionId;
        final ResponseDefinition response = wireMockServer.stubFor(
                put(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedSubmission))
                                .withHeader("Content-Type", MediaType.APPLICATION_JSON_VALUE)
                                .withStatus(200))).getResponse();


        // When
        final Submission submission = underTest.sendSubmission(submitSubmission);

        // Then
        assertNotNull(submission);
        assertThat(submission.getSubmissionId(), is(submissionId));
        assertThat(submission.getEncryptedData(), is(encryptedData));
        assertThat(submission.getEncryptedMetadata(), is(encryptedMetadata));

        verify(putRequestedFor(urlEqualTo(urlPath)).withHeader("Authorization", equalTo("Bearer abc")));
        assertThat(response.getHeaders().getContentTypeHeader().mimeTypePart(), is("application/json"));
        assertEquals(response.getStatus(), HttpStatus.OK.value());
    }

    @Test
    void pollAvailableSubmissions() throws JsonProcessingException {

        final int offset = 0;
        final int limit = 10;
        final var destinationId = UUID.randomUUID();

        final Set<SubmissionForPickup> allSubmissions = generateSubmission(destinationId, 100);

        final Set<SubmissionForPickup> pagedSubmissions = allSubmissions.stream()
                .skip(offset)
                .limit(limit)
                .collect(Collectors.toSet());

        final SubmissionsForPickup expectedSubmissions = new SubmissionsForPickup();
        expectedSubmissions.setSubmissions(pagedSubmissions);
        expectedSubmissions.setCount(pagedSubmissions.size());
        expectedSubmissions.setOffset(offset);
        expectedSubmissions.setTotalCount(allSubmissions.size());

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/submissions?destinationId=" + destinationId + "&limit=" + limit + "&offset=" + offset;
        final ResponseDefinition response = wireMockServer.stubFor(
                get(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedSubmissions))
                                .withHeader("Content-Type", MediaType.APPLICATION_JSON_VALUE)
                                .withStatus(200))).getResponse();

        // When
        final SubmissionsForPickup submissions = underTest.pollAvailableSubmissions(destinationId, offset, limit);

        // Then
        assertNotNull(submissions);
        assertThat(submissions.getOffset(), is(offset));
        assertThat(submissions.getSubmissions().size(), is(limit));
        assertThat(submissions.getTotalCount(), is(allSubmissions.size()));
    }

    @Test
    void getSubmission() throws JsonProcessingException {
        // Given
        final var submissionId = UUID.randomUUID();

        final var authToken = new OAuthToken();
        authToken.setAccessToken("abc");
        authToken.setExpiresIn(1800);

        final var expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(submissionId);

        when(authServiceMock.getCurrentToken()).thenReturn(authToken);

        final var urlPath = "/v1/submissions/" + submissionId;
        final ResponseDefinition response = wireMockServer.stubFor(
                get(urlEqualTo(urlPath))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(expectedSubmission))
                                .withHeader("Content-Type", MediaType.APPLICATION_JSON_VALUE)
                                .withStatus(200))).getResponse();

        final Submission submission = underTest.getSubmission(submissionId);

        assertNotNull(submission);
        assertThat(submission.getSubmissionId(), is(expectedSubmission.getSubmissionId()));
    }

    private Set<SubmissionForPickup> generateSubmission(final UUID destinationId, final int count){
        final Set<SubmissionForPickup> submissions = new HashSet<>(count);
        for (int i = 0; i < count; i++) {
            final SubmissionForPickup submission = new SubmissionForPickup();
            submission.setDestinationId(destinationId);
            submission.setSubmissionId(UUID.randomUUID());
            submissions.add(submission);
        }
        return submissions;
    }

}